"""

     esbmtk.connections

     Classes which handle the connections and fluxes between esbmtk objects
     like Reservoirs, Sources, and Sinks.

     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# from pint import UnitRegistry
from __future__ import annotations
from numbers import Number
from nptyping import *
from typing import *
from numpy import array, set_printoptions, arange, zeros, interp, mean
from pandas import DataFrame
from copy import deepcopy, copy
from time import process_time

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

import logging
import time
import builtins

set_printoptions(precision=4)
from .utility_functions import map_units
from .processes import *
from .esbmtk import *
from esbmtk import Q_
from .utility_functions import check_for_quantity


class Connect(esbmtkBase):
    """Two reservoirs connect to each other via at least one flux. This
     module creates the connecting flux and creates a connector object
     which stores all connection properties.

     For simple connections, the type flux type is derived implcitly from the specified parameters.
     For complex connections, the flux type must be set explicitly. See the examples below:

     Parameters:
         - name: A string which determines the name of this object. Optional, if not provided
           the connection name will be derived as "C_Source_2_Sink"
         - source: An object handle for a Source or Reservoir
         - sink: An object handle for a Sink or Reservoir
         - rate: A quantity (e.g., "1 mol/s"), optional
         - delta: The isotope ratio, optional
         - ref_reservoirs: Reservoir or flux reference
         - alpha: A fractionation factor, optional
         - id: A string wich will become part of the object name, optional
         - plot: "yes" or "no", defaults to "yes"
         - signal: An object handle of signal, optional
         - pl: A list of process objects, optional
         - ctype: connection type, optional, this allows to scale a flux in response to other
           reservoirs and fluxes
         - bypass :str optional defaults to "None" see scale with flux
         - save_flux_data: bool, defaults to model setting, unles specified

     Connection Types:
     -----------------
     Basic Connections (the advanced ones are below):

     - If both =rate= and =delta= are given, the flux is treated as a
        fixed flux with a given isotope ratio. This is usually the case for
        most source objects (they can still be affected by a signal, see
        above), but makes little sense for reservoirs and sinks.

     - If both the =rate= and =alpha= are given, the flux rate is fixed
       (subject to any signals), but the isotopic ratio of the output
       flux depends on the isotopic ratio of the upstream reservoir
       plus any isotopic fractionation specified by =alpha=. This is
       typically the case for fluxes which include an isotopic
       fractionation (i.e., pyrite burial). This combination is not
       particularly useful for source objects.

     - If the connection specifies only =delta= the flux is treated as a
       variable flux which is computed in such a way that the reservoir
       maintains steady state with respect to it's mass.

     - If the connection specifies only =rate= the flux is treated as a
       fixed flux which is computed in such a way that the reservoir
       maintains steady state with respect to it's isotope ratio.

     Examples of Basic Connections
     -----------------------------

     Connecting a Source to a Reservoir
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

     Unless you use a Signal, a source typically provides a steady stream with a given isotope ratio (if used)

     Example::

        Connect(source =  Source,
                sink = downstrean reservoir,
                rate = "1 mol/s",
                delta = optional,
                signal = optional, see the signal documentation)

     Connecting a Reservoir to Sink or another Reservoir
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

     Here we can distinguish between cases where we use fixed flux, or a flux that reacts to in some way to the
     upstream reservoir (see the Reservoir to Reservoir section for a more complete treatment):

     Fixed outflux, with no isotope fractionation

     Example::

          Connect(source =  upstream reservoir,
                sink = Sink,
                rate = "1 mol/s",)

     Fixed outflux, with isotope fractionation

     Example::

          Connect(source =  upstream reservoir,
                sink = Sink,
                alpha = -28,
                rate = "1 mol/s",)

     Advanced Connections
     --------------------

     You can aditionally define connection properties via the ctype
     keyword. This requires additional keyword parameters. The following values are
     recognized

     ctype = "scale_with_flux"
     ~~~~~~~~~~~~~~~~~~~~~~~~~

     This will scale a flux relative to another flux:

     Example::

         Connect(source =  upstream reservoir,
                sink = downstream reservoir,
                ctype = "scale_with_flux",
                ref_flux = flux handle,
                scale = a scaling factor, optional, defaults to 1
                bypass :str = "source"/"sink"
                )

    if bypass is set the flux will not affect the upstream or
    downstream reservoir. This is a more generalized approach than the
    virtual flux type. This is useful to describe reactions which
    split into two elements (i.e., deprotonation), or combine fluxes
    from two reservoirs (say S, and O) into a reservoir which records
    SO4.

    ctype = "virtual_flux"
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    This will scale a flux relative to another flux, similar to
    scaleflux. However, in this case, the flux will only affect the
    sink, and not the source. This is useful to describe reactions which
    split into two elements (i.e., deprotonation).

    Example::

         Connect(source =  upstream reservoir,
                sink = downstream reservoir,
                ctype = "scale_with_flux",
                ref_flux = flux handle,
                scale = a scaling factor, optional, defaults to 1
                )

     ctype = "scale_with_mass" and "scale_with_concentration"
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

     This will scale a flux relative to the mass or concentration of a reservoir

     Example::

         Connect(source =  upstream reservoir,
                sink = downstream reservoir,
                ctype = "scale_with_mass",
                ref_reservoirs = reservoir handle,
                scale = a scaling factor, optional, defaults to 1
    )

     ctype = "scale_relative_to_multiple_reservoirs"
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

     This process scales the flux as a function one or more reservoirs
     or constants which describes the
     strength of relation between the reservoir concentrations and
     the flux scaling

     F = C1 * C2 * k

     where Ci denotes the concentrartion in one  or more reservoirs, k is one
     or more constants.

     Example::

         Connect(source =  upstream reservoir,
                sink = downstream reservoir,
                ctype = "scale_relative_to_multiple_reservoirs"
                ref_reservoirs = [r1, r2, k etc] # you must provide at least one
                scale = a scaling factor, optional, defaults to 1
    )

     scale is an overall scaling factor.


     ctype = "flux_balance"
     ~~~~~~~~~~~~~~~~~~~~~~

    This type can be used to express equilibration fluxes
    between two reservoirs. This connection type, takes three parameters:

    - =left= is a list which can contain constants and/or reservoirs. The
      list must contain at least one valid element. All elements in this
      list will be multiplied with each other. E.g. if we have a list
      with one constant and one reservoir, the reservoir concentration
      will be multiplied with the constant. If we have two reservoirs,
      the respective reservoir concentrations will be multiplied with
      each other.
    - =right= similar to =left= The final flux rate will be computed as
      the difference between =left= and =right=
    - =k_value= a constant which will be multiplied with the difference
      between =left=and =right=

     Example::

         Connect(source=R_CO2,         # target of flux
                 sink=R_HCO3,          # source of flux
                 rate="1 mol/s",       # flux rate
                 ctype="flux_balance", # connection type
                 scale=1,            # global scaling factor, optional
                 left=[K1, R_CO2],     # where K1 is a constant
                 right=[R_HCO3, R_Hplus])


     Useful methods in this class
     ----------------------------
     The following methods might prove useful

      - info() will provide a short description of the connection objects.
      - list_processes() which will list all the processes which are associated with this connection.
      - update() which allows you to update connection properties after the connection has been created

    """

    def __init__(self, **kwargs):
        """The init method of the connector obbjects performs sanity checks e.g.:
               - whether the reservoirs exist
               - correct flux properties (this will be handled by the process object)
               - whether the processes do exist (hmmh, that implies that the optional processes do get registered with the model)
               - creates the correct default processes
               - and connects the reservoirs

        see the class documentation for details and examples

        """

        from . import ureg, Q_

        # provide a dict of all known keywords and their type
        self.lkk: Dict[str, any] = {
            "name": str,
            "id": str,
            "source": (Source, Reservoir, GasReservoir),
            "sink": (Sink, Reservoir, GasReservoir),
            "delta": (Number, str),
            "rate": (str, Number, Q_),
            "pl": list,
            "alpha": (Number, str),
            "species": Species,
            "ctype": str,
            "ref_reservoirs": (Reservoir, GasReservoir, str, list),
            "ref_flux": (Flux, str, list),
            "ratio": Number,
            "scale": (Number, Q_, str),
            "ref_value": (str, Number, Q_),
            "k_value": (Number, str, Q_),
            "a_value": Number,
            "b_value": Number,
            "left": (list, Number, Reservoir, GasReservoir),
            "right": (list, Number, Reservoir, GasReservoir),
            "plot": str,
            "groupname": bool,
            "register": any,
            "signal": (Signal, str),
            "bypass": (str, Reservoir, GasReservoir),
            "isotopes": bool,
            "solubility": Number,
            "area": Number,
            "piston_velocity": Number,
            "function_ref": any,
            "save_flux_data": (bool, str),
        }

        # provide a list of absolutely required keywords
        self.lrk: list = ["source", "sink"]

        # list of default values if none provided
        self.lod: Dict[any, any] = {
            "id": "",
            "plot": "yes",
            "ctype": "None",
            "delta": "None",
            "alpha": "None",
            "rate": "None",
            "k_value": "None",
            "scale": 1,
            "signal": "None",
            "groupname": False,
            "bypass": "None",
            "name": "None",
            "isotopes": False,
            "ref_reservoirs": "None",
            "ref_flux": "None",
            "register": "None",
            "function_ref": "None",
            "save_flux_data": "None",
        }

        # validate and initialize instance variables
        self.__initerrormessages__()

        self.bem.update(
            {
                "k_concentration": "a number",
                "k_mass": "a number",
                "k_value": "a number",
                "scale": "a number or Quantity",
                "a_value": "a number",
                "ref_value": "a number, string, or quantity",
                "b_value": "a number",
                "name": "a string",
                "id": "a string",
                "plot": "a string",
                "left": "Number, list or Reservoir",
                "right": "Number, list or Reservoir",
                "signal": "Signal Handle",
                "groupname": "True or False",
                "bypass": "source/sink",
                "function_ref": "A function",
            }
        )

        self.drn = {
            "alpha": "_alpha",
            "rate": "_rate",
            "delta": "_delta",
        }

        self.__validateandregister__(kwargs)

        # if kwargs["id"] != "None":
        #    self.name = self.name + f"_{self.id}"
        if "pl" in kwargs:
            self.lop: list[Process] = self.pl
        else:
            self.lop: list[Process] = []

        if self.signal != "None":
            self.lop.append(self.signal)

        # if we have a signal, save flux data
        if self.signal != "None":
            self.save_flux_data = True
        # else if save_flux_data is unsepcified, use model default
        elif self.save_flux_data == "None":
            self.save_flux_data = self.source.sp.mo.save_flux_data

        # if no reference reservoir is specified, default to the upstream
        # reservoir
        if self.ref_reservoirs == "None":
            self.ref_reservoirs = kwargs["source"]

        # decide if this connection needs isotope calculations
        if self.source.isotopes or self.sink.isotopes:
            self.isotopes = True

        # legacy names
        self.influx: int = 1
        self.outflux: int = -1
        # self.n = self.name
        self.mo = self.source.sp.mo
        self.p = 0  # the default process handle
        self.r1: (Process, Reservoir) = self.source
        self.r2: (Process, Reservoir) = self.sink

        self.get_species(self.r1, self.r2)  #
        self.mo: Model = self.sp.mo  # the current model handle
        self.lof: list[Flux] = []  # list of fluxes in this connection
        # get a list of all reservoirs registered for this species
        self.lor: list[Reservoir] = self.mo.lor

        # make sure scale is a number in model units

        if self.scale == "None":
            self.scale = 1.0

        if isinstance(self.scale, str):
            self.scale = Q_(self.scale)

        if isinstance(self.scale, Q_):
            # test what type of Quantity we have
            if self.scale.check(["volume]/[time"]):  # flux
                self.scale = self.scale.to(self.mo.r_unit)
            elif self.scale.check(["mass] / [time"]):  # flux
                self.scale = self.scale.to(self.mo.f_unit)
            elif self.scale.check("[mass]/[volume]"):  # concentration
                self.scale = self.scale.to(self.mo.c_unit)
            else:
                ValueError(f"No conversion to model units for {self.scale} specified")

        # if sink and source a regular, the name will be simply C_S_2_S
        # if we deal with ReservoirGroups we need to reflect this in the
        # connection name

        self.__set_name__()  # get name of connection

        self.__create_flux__()  # Source/Sink/Regular

        self.__set_process_type__()  # derive flux type and create flux(es)

        # print(f"mo.reg = {self.mo.register}, slf reg = {self.register}")
        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name__()  # register connection in namespace

        self.source.loc.add(self)  # register connector with reservoir
        self.sink.loc.add(self)  # register connector with reservoir
        self.mo.loc.add(self)  # register connector with model

        # This should probably move to register fluxes
        self.__register_process__()

        # if self.register == "None":
        #     print(f"Created connection {self.name}")
        # else:
        #     print(f"Created connection {self.register.name}.{self.name}")

        logging.info(f"Created {self.full_name}")

    def __set_name__(self):
        """set connection name if not explicitly provided"""

        if self.name == "None":
            self.name = f"C_{self.source.name}_2_{self.sink.name}"

            if self.id == "None" or self.id == "":
                pass
            else:
                self.name = f"{self.name}_{self.id}"

        if self.register == "None":
            self.full_name = f"{self.mo.name}.{self.name}"
        else:
            self.full_name = f"{self.register.full_name}.{self.name}"

        self.base_name = self.full_name
        self.n = self.name

        # if self.groupname: # if we are part of a group connection
        #     self.name = f"C_{self.source.name}_2_{self.sink.name}"
        #     if self.id == "None" or self.id == "":
        #         pass
        #     else:
        #         self.name = f"{self.name}_{self.id}"

        #     self.full_name = f"{self.register.full_name}.{self.name}"
        # else:
        #     if self.name == "None":
        #         self.name = f"C_{self.source.name}_2_{self.sink.name}"
        #     else:
        #         self.name = f"{self.name}"

        #     if self.id == "None" or self.id == "":
        #         pass
        #     else:
        #         self.name = f"{self.name}_{self.id}"

        #     self.full_name = self.name

        # self.base_name = self.full_name
        # self.n = self.name

    def update(self, **kwargs):
        """Update connection properties. This will delete existing processes
        and fluxes, replace existing key-value pairs in the
        self.kwargs dict, and then re-initialize the connection.

        """

        raise NotImplementedError
        self.__delete_process__()
        self.__delete_flux__()
        self.kwargs.update(kwargs)
        self.__set__name__()  # get name of connection
        self.__init_connection__(self.kwargs)

    def get_species(self, r1, r2) -> None:
        """In most cases the species is set by r2. However, if we have
        backward fluxes the species depends on the r2

        """
        # print(f"r1 = {r1.n}, r2 = {r2.n}")
        if isinstance(self.r1, Source):
            self.r = r1
        else:  # in this case we do have an upstream reservoir
            self.r = r2

        # test if species was explicitly given
        if "species" in self.kwargs:  # this is a quick fix only
            self.sp = self.kwargs["species"]
        else:
            self.sp = self.r.sp  # get the parent species

    def __create_flux__(self) -> None:
        """Create flux object, and register with reservoir and global namespace"""

        # test if default arguments present
        if self.delta == "None":
            d = 0
        else:
            d = self.delta

        if self.rate == "None":
            r = f"0 {self.sp.mo.f_unit}"
            # self._rate = r
        else:
            r = self.rate

        # flux name
        if self.register == "None":
            if self.id == "None" or self.id == "":
                n = f"{self.r1.n}_2_{self.r2.n}_F"
            else:
                n = f"{self.r1.n}_2_{self.r2.n}_{self.id}_F"
        else:
            if self.id == "None" or self.id == "":
                n = f"{self.r1.n}_2_{self.r2.n}_F"
            else:
                n = f"{self.id}_F"
        # else:
        #    n = "F_" + self.r1.full_name + "_2_" + self.r2.full_name

        # derive flux unit from species obbject
        funit = self.sp.mu + "/" + str(self.sp.mo.bu)  # xxx

        self.fh = Flux(
            name=n,  # flux name
            species=self.sp,  # Species handle
            delta=d,  # delta value of flux
            rate=r,  # flux value
            plot=self.plot,  # display this flux?
            register=self,  # is this part of a group?
            # register=self.register,  # is this part of a group?
            isotopes=self.isotopes,
            id=self.id,
            save_flux_data=self.save_flux_data,
        )

        # register flux with its reservoirs
        if isinstance(self.r1, Source):
            # add the flux name direction/pair
            self.r2.lio[self.fh] = self.influx
            # add the handle to the list of fluxes
            self.r2.lof.append(self.fh)
            # register flux and element in the reservoir.
            self.__register_species__(self.r2, self.r1.sp)

        elif isinstance(self.r2, Sink):
            # add the flux name direction/pair
            self.r1.lio[self.fh] = self.outflux
            # add flux to the upstream reservoir
            self.r1.lof.append(self.fh)
            # register flux and element in the reservoir.
            self.__register_species__(self.r1, self.r2.sp)

        elif isinstance(self.r1, Sink):
            raise NameError(
                "The Sink must be specified as a destination (i.e., as second argument"
            )

        elif isinstance(self.r2, Source):
            raise NameError("The Source must be specified as first argument")

        else:  # this is a regular connection
            # add the flux name direction/pair
            self.r1.lio[self.fh] = self.outflux
            # add the flux name direction/pair
            self.r2.lio[self.fh] = self.influx
            # add flux to the upstream reservoir
            self.r1.lof.append(self.fh)
            # add flux to the downstream reservoir
            self.r2.lof.append(self.fh)
            self.__register_species__(self.r1, self.r1.sp)
            self.__register_species__(self.r2, self.r2.sp)

        self.lof.append(self.fh)

    def __register_species__(self, r, sp) -> None:
        """Add flux to the correct element dictionary"""
        # test if element key is present in reservoir
        if sp.eh in r.doe:
            # add flux handle to dictionary list
            r.doe[sp.eh].append(self.fh)
        else:  # add key and first list value
            r.doe[sp.eh] = [self.fh]

    def __register_process__(self) -> None:
        """Register all flux related processes"""

        # first test if we have a signal in the list. If so,
        # remove signal and replace with process

        p_copy = copy(self.lop)
        for p in p_copy:  # loop over process list if provided during init
            if isinstance(p, Signal):
                self.lop.remove(p)
                if p.ty == "addition":
                    # create AddSignal Process object
                    n = AddSignal(
                        name=p.n + "_addition_process",
                        reservoir=self.r,
                        flux=self.fh,
                        lt=p.data,
                        register=self.fh,
                    )
                    self.lop.insert(0, n)  # signals must come first
                    logging.debug(f"Inserting {n.n} in {self.name} for {self.r.n}")
                elif p.ty == "multiplication":
                    # create AddSignal Process object
                    n = MultiplySignal(
                        name=p.n + "_multiplication_process",
                        reservoir=self.r,
                        flux=self.fh,
                        lt=p.data,
                        register=self.fh,
                    )
                    self.lop.insert(len(self.lop), n)  # multiplaction should come last
                    logging.debug(f"Inserting {n.n} in {self.name} for {self.r.n}")
                else:
                    raise ValueError(f"Signal type {p.ty} is not defined")

        # ensure that processes are in the correct order
        self.__move_process_to_top_of_queue__(self.lop, MultiplySignal)
        self.__move_process_to_top_of_queue__(self.lop, AddSignal)
        self.__move_process_to_top_of_queue__(self.lop, ScaleFlux)
        self.__move_process_to_top_of_queue__(self.lop, VarDeltaOut)
        self.__move_process_to_end_of_queue__(self.lop, SaveFluxData)

        # nwo we can register everythig on lop
        i = 0
        for p in self.lop:
            if isinstance(p, (MultiplySignal)) and i > 0:
                p.__execute__ = p.__multiply_with_flux_fa__
                p.__get_process_args__ = p.__get_process_args_fa__
            if isinstance(p, (AddSignal)) and i > 0:
                print(f"Found Add signal")
                p.__execute__ = p.__add_with_fa__
                p.__get_process_args__ = p.__get_process_args_fa__

            p.__register__(self.r, self.fh)
            i += 1

    def __move_process_to_top_of_queue__(self, lop: list, ptype: any) -> None:
        """Return a copy of lop where ptype has been moved to the top of lop"""
        p_copy = copy(lop)
        for p in p_copy:  # loop over process list if provided during init
            if isinstance(p, ptype):
                lop.remove(p)
                lop.insert(0, p)  # insert at top

    def __move_process_to_end_of_queue__(self, lop: list, ptype: any) -> None:
        """Return a copy of lop where ptype has been moved to the top of lop"""
        p_copy = copy(lop)
        for p in p_copy:  # loop over process list if provided during init
            if isinstance(p, ptype):
                lop.remove(p)
                lop.append(p)  # insert at end

    def __set_process_type__(self) -> None:
        """Deduce flux type based on the provided flux properties. The method calls the
        appropriate method init routine
        """

        if isinstance(self.r1, Source):
            self.r = self.r2
        else:
            self.r = self.r1

        # if signal is provided but rate is omitted
        if self.signal != "None" and self.rate == "None":
            self._rate = "0 mmol/y"

        # if connection type is not set explicitly
        if self.ctype == "None" or self.ctype == "Regular":
            # set the fundamental flux type based on the flux arguments given
            if self.delta != "None" and self.rate != "None":
                # self.__vardeltaout__()
                pass  # if delta and rate are specified, do nothing
            # variable flux with fixed delta
            elif self.delta != "None":  # if delta is set
                self.__passivefluxfixeddelta__()
            elif self.rate != "None":  # if rate is set
                if self.delta != "None" and not isinstance(self.source, Sink):
                    self.__vardeltaout__()  # variable delta with fixed flux
            else:  # if neither are given -> default varflux type
                self._delta = 0
                self.__passiveflux__()

        elif self.ctype == "scale_to_input":
            self.__scale_to_input__()
        elif self.ctype == "flux_diff":
            self.__vardeltaout__()
            self.__flux_diff__()
        elif self.ctype == "scale_with_flux":
            self.__scaleflux__()
        elif self.ctype == "weathering":
            self.__rateconstant__()
        elif self.ctype == "virtual_flux":
            self.__virtual_flux__()
            # self.__vardeltaout__()
        elif self.ctype == "copy_flux":
            self.__scaleflux__()
            # self.__vardeltaout__()
        elif self.ctype == "scale_with_mass":
            self.__rateconstant__()
        elif self.ctype == "scale_with_concentration":
            self.__rateconstant__()
        elif self.ctype == "scale_relative_to_multiple_reservoirs":
            self.__rateconstant__()
        elif self.ctype == "flux_balance":
            self.__rateconstant__()
        elif self.ctype == "react_with":
            self.__reaction__()
        elif self.ctype == "monod_type_limit":
            # self.__vardeltaout__()
            self.__rateconstant__()
        elif self.ctype == "manual":
            pass
        else:
            print(f"Connection Type {self.ctype} is unknown")
            raise ValueError(f"Unknown connection type {self.ctype}")

        # Set optional flux processes
        if self.alpha != "None":
            self.__alpha__()  # Set optional flux processes
            # self.__vardeltaout__()

        # check if flux should bypass any reservoirs
        if self.bypass == "source":
            self.source.lof.remove(self.fh)
        elif self.bypass == "sink":
            self.sink.lof.remove(self.fh)
            # print(f"removing {self.fh.full_name} from {self.sink.full_name} lof")
        elif self.bypass == "None":
            pass
        else:
            raise ValueError(f"bypass must be None/source/sink but not {self.bypass}")

        if self.save_flux_data:
            ph = SaveFluxData(
                name=f"{self.fh.full_name}_Pfd",
                flux=self.fh,
                register=self.fh,
            )
            self.lop.append(ph)

    def __passivefluxfixeddelta__(self) -> None:
        """Just a wrapper to keep the if statement manageable"""

        ph = PassiveFlux_fixed_delta(
            name="Pfd",
            reservoir=self.r,
            flux=self.fh,
            register=self.fh,
            delta=self.delta,
        )  # initialize a passive flux process object
        self.lop.append(ph)

    def __vardeltaout__(self) -> None:
        """Unlike a passive flux, this process sets the output flux from a
        reservoir to a fixed value, but the isotopic ratio of the
        output flux will be set equal to the isotopic ratio of the
        upstream reservoir.

        """

        ph = VarDeltaOut(
            name="Pvdo",
            reservoir=self.source,
            flux=self.fh,
            register=self.fh,
            rate=self.rate,
        )
        self.lop.append(ph)

    def __scaleflux__(self) -> None:
        """Scale a flux relative to another flux"""

        if not isinstance(self.ref_flux, Flux):
            raise ValueError("Scale reference must be a flux")

        if self.k_value != "None":
            self.scale = self.k_value
            print(f"\n Warning: use scale instead of k_value for scaleflux type\n")

        ph = ScaleFlux(
            name="PSF",
            source=self.source,
            reservoir=self.r,
            flux=self.fh,
            register=self.fh,
            scale=self.scale,
            ref_flux=self.ref_flux,
            delta=self.delta,
        )
        self.lop.append(ph)

        # if self.bypass == "source":
        #     self.source.lof.remove(self.fh)
        # elif self.bypass == "sink":
        #     self.sink.lof.remove(self.fh)
        #     print(f"removing {self.fh.full_name} from {self.sink.full_name} lof")
        # elif self.bypass == "None":
        #     pass
        # else:
        #     raise ValueError(f"bypass must be None/source/sink but not {self.bypass}")

    def __virtual_flux__(self) -> None:
        """Create a virtual flux. This is similar to __scaleflux__, however the new flux
        will only affect the sink, and not the source.

        """

        if self.k_value != "None":
            self.scale = self.k_value
            print(
                f"\n Warning: use scale instead of k_value for scale relative to multiple reservoirs\n"
            )

        if not isinstance(self.kwargs["ref_reservoirs"], Flux):
            raise ValueError("Scale reference must be a flux")

        ph = ScaleFlux(
            name="PSFV",
            reservoir=self.r,
            flux=self.fh,
            register=self.fh,
            scale=self.scale,
            ref_flux=self.ref_flux,
            delta=self.delta,
        )
        self.lop.append(ph)

        # this flux must not affect the source reservoir
        self.r.lof.remove(self.fh)

    def __flux_diff__(self) -> None:
        """Scale a flux relative to the difference between
        two fluxes

        """

        if self.k_value != "None":
            self.scale = self.k_value
            print(
                f"\n Warning: use scale instead of k_value for scale relative to multiple reservoirs\n"
            )

        if not isinstance(self.kwargs["ref_reservoirs"], list):
            raise ValueError("ref must be a list")

        ph = FluxDiff(
            name="PSF",
            reservoir=self.r,
            flux=self.fh,
            register=self.fh,
            scale=self.scale,
            ref_reservoirs=self.ref_reservoirs,
        )
        self.lop.append(ph)

    def __reaction__(self) -> None:
        """Just a wrapper to keep the if statement manageable"""

        if not isinstance(self.ref_reservoirs, (Flux)):
            raise ValueError("Scale reference must be a flux")

        if self.k_value != "None":
            self.scale = self.k_value
            print(f"\n Warning: use scale instead of k_value for reaction type\n")

        ph = Reaction(
            name="RF",
            reservoir=self.r,
            flux=self.fh,
            register=self.fh,
            scale=self.scale,
            ref_reservoirs=self.ref_reservoirs,
        )

        # we need to make sure to remove the flux referenced by
        # react_with is removed from the list of fluxes in this
        # reservoir. This should probably move in to the React Class
        self.r2.lof.remove(self.ref_reservoirs)
        self.lop.append(ph)

    def __passiveflux__(self) -> None:
        """Just a wrapper to keep the if statement manageable"""

        ph = PassiveFlux(
            name="_PF",
            reservoir=self.r,
            register=self.fh,
            flux=self.fh,
            scale=self.scale,
        )  # initialize a passive flux process object
        self.lop.append(ph)  # add this process to the process list

    def __scale_to_input__(self) -> None:
        """Just a wrapper to keep the if statement manageable"""

        ph = ScaleRelativeToInputFluxes(
            name="_SRTIF",
            reservoir=self.r,
            register=self.fh,
            flux=self.fh,
            scale=self.scale,
        )  # initialize a passive flux process object
        self.lop.append(ph)  # add this process to the process list

    def __alpha__(self) -> None:
        """Just a wrapper to keep the if statement manageable"""

        ph = Fractionation(
            name="_Pa",
            reservoir=self.r,
            flux=self.fh,
            register=self.fh,
            alpha=self.kwargs["alpha"],
        )
        self.lop.append(ph)  #

    def __rateconstant__(self) -> None:
        """Add rate constant type process"""

        from . import ureg, Q_

        # this process requires that we use the vardeltaout process
        if self.mo.m_type != "mass_only":
            # self.__vardeltaout__()
            pass

        if self.ctype == "scale_with_mass":
            if self.k_value != "None":
                self.scale = self.k_value
                print(
                    f"\n Warning: use scale instead of k_value for scale with mass type\n"
                )

            self.scale = map_units(self.scale, self.mo.m_unit)
            ph = ScaleRelativeToMass(
                name="_PkM",
                reservoir=self.ref_reservoirs,
                flux=self.fh,
                register=self.fh,
                scale=self.scale,
            )

        elif self.ctype == "scale_with_concentration":

            if self.k_value != "None":
                self.scale = self.k_value
                print(
                    f"\n Warning: use scale instead of k_value for scale with concentration type\n"
                )

            self.scale = map_units(
                self.scale, self.mo.c_unit, self.mo.f_unit, self.mo.r_unit
            )
            # print(
            #    f"Registering PKC with ref = {self.ref.full_name}, scale = {self.scale}"
            # )

            ph = ScaleRelativeToConcentration(
                name="PkC",
                reservoir=self.ref_reservoirs,
                flux=self.fh,
                register=self.fh,
                scale=self.scale,
            )
            # print(f"Process Name {ph.full_name}")

        elif self.ctype == "weathering":

            ph = weathering(
                name="Pw",
                source=self.source,
                reservoir=self.r,
                flux=self.fh,
                register=self.fh,
                scale=self.scale,
                reservoir_ref=self.reservoir_ref,
                ex=self.ex,
                pco2_0=self.pco2_0,
                f_0=self.rate,
                delta=self.delta,
            )

        elif self.ctype == "scale_relative_to_multiple_reservoirs":

            if self.k_value != "None":
                self.scale = self.k_value
                print(
                    f"\n Warning: use scale instead of k_value for scale relative to multiple reservoirs\n"
                )

            self.scale = map_units(
                self.scale, self.mo.c_unit, self.mo.f_unit, self.mo.r_unit
            )
            ph = ScaleRelative2otherReservoir(
                name="PkC",
                reservoir=self.source,
                ref_reservoirs=self.ref_reservoirs,
                flux=self.fh,
                register=self.fh,
                scale=self.scale,
            )

        elif self.ctype == "flux_balance":
            self.k_value = map_units(
                self.k_value, self.mo.c_unit, self.mo.f_unit, self.mo.r_unit
            )
            ph = Flux_Balance(
                name="_Pfb",
                reservoir=self.source,
                left=self.left,
                right=self.right,
                flux=self.fh,
                register=self.fh,
                k_value=self.k_value,
            )

        elif self.ctype == "monod_ctype_limit":
            self.ref_value = map_units(self.ref_value, self.mo.c_unit)
            ph = Monod(
                name="_PMonod",
                reservoir=self.ref_reservoirs,
                flux=self.fh,
                register=self.fh,
                ref_value=self.ref_value,
                a_value=self.a_value,
                b_value=self.b_value,
            )

        else:
            raise ValueError(
                f"This should not happen,and points to a keywords problem in {self.name}"
            )

        # print(f"adding {ph.name} to {self.name}")
        self.lop.append(ph)

    def info(self, **kwargs) -> None:
        """Show an overview of the object properties.
        Optional arguments are
        index  :int = 0 this will show data at the given index
        indent :int = 0 indentation

        """
        off: str = "  "
        if "index" not in kwargs:
            index = 0
        else:
            index = kwargs["index"]

        if "indent" not in kwargs:
            indent = 0
            ind = ""
        else:
            indent = kwargs["indent"]
            ind = " " * indent

        # print basic data bout this Connection
        print(f"{ind}{self.__str__(kwargs)}")

        print(f"{ind}Fluxes:")
        for f in sorted(self.lof):
            f.info(indent=indent, index=index)

    def __delete_process__(self) -> None:
        """Updates to the connection properties may change the connection type and thus
        the processes which are associated with this connection. We thus have to
        first delete the old processes, before we re-initialize the connection

        """

        # identify which processes we need to delete
        # unregister process from connection.lop, reservoir.lop, flux.lop, model.lmo
        # delete process from global name space if present

        lop = copy(self.lop)

        for p in lop:
            for f in self.lof:
                if isinstance(f.register, ConnectionGroup):
                    # remove from Connection group list of model objects
                    self.register.lmo.remove(f)
                else:
                    self.r1.lop.remove(p)
                    self.fh.lop.remove(p)
                    self.lop.remove(p)
                    self.r1.mo.lmo.remove(p.n)
                    del p

    def __delete_flux__(self) -> None:
        """Updates to the connection properties may change the connection type and thus
        the processes which are associated with this connection. We thus have to
        first delete the old flux, before we re-initialize the connection

        """

        # identify which processes we need to delete
        # unregister process from connection.lop, reservoir.lop, flux.lop, model.lmo
        # delete process from global name space if present

        lof = copy(self.lof)
        for f in lof:
            if isinstance(f.register, ConnectionGroup):
                # remove from Connection group list of model objects
                self.register.lmo.remove(f)
            else:
                self.r1.lof.remove(f)
                self.lof.remove(f)
                self.r1.mo.lmo.remove(f.n)
                del f

    # ---- Property definitions to allow for connection updates --------
    """ Changing the below properties requires that we delete all
    associated objects (processes), and determines the new flux type,
    and initialize/register these with the connection and model.
    We also have to update the keyword arguments as these are used
    for the log entry

    """

    # ---- alpha ----
    @property
    def alpha(self) -> Number:
        return self._alpha

    @alpha.setter
    def alpha(self, a: Number) -> None:
        self.__delete_process__()
        self.__delete_flux__()
        self._alpha = a
        self.kwargs["alpha"] = a
        self.__set_process_type__()  # derive flux type and create flux(es)
        self.__register_process__()

    # ---- rate  ----
    @property
    def rate(self) -> Number:
        return self._rate

    @rate.setter
    def rate(self, r: str) -> None:
        from . import ureg, Q_

        self.__delete_process__()
        self.__delete_flux__()
        self._rate = Q_(r).to(self.mo.f_unit)
        self.kwargs["rate"] = r
        self.__create_flux__()  # Source/Sink/Regular
        self.__set_process_type__()  # derive flux type and create flux(es)
        self.__register_process__()

    # ---- delta  ----
    @property
    def delta(self) -> Number:
        return self._delta

    @delta.setter
    def delta(self, d: Number) -> None:
        self.__delete_process__()
        self.__delete_flux__()
        self._delta = d
        self.kwargs["delta"] = d
        self.__create_flux__()  # Source/Sink/Regular
        self.__set_process_type__()  # derive flux type and create flux(es)
        self.__register_process__()


class Connection(Connect):
    """Alias for the Connect class"""


class ConnectionGroup(esbmtkBase):
    """Name:

        ConnectionGroup

        Connect reservoir/sink/source groups when at least one of the
        arguments is a reservoirs_group object. This method will
        create regular connections for each matching species.

        Use the connection.update() method to fine tune connections
        after creation

    Example::

        ConnectionGroup(source =  upstream reservoir / upstream reservoir group
           sink = downstrean reservoir / downstream reservoirs_group
           delta = defaults to zero and has to be set manually
           alpha =  defaults to zero and has to be set manually
           rate = shared between all connections
           ref_reservoirs = shared between all connections
           ref_flux = shared between all connections
           species = list, optional, if present, only these species will be connected
           ctype = needs to be set for all connections. Use "Regular"
                   unless you require a specific connection type
           pl = [list]) process list. optional, shared between all connections
           id = optional identifier, passed on to individual connection
           plot = "yes/no" # defaults to yes, shared between all connections
           save_flux_data = True/False, use model default if not set
        )

        ConnectionGroup(
                  source=OM_Weathering,
                  sink=Ocean,
                  rate={DIC: f"{OM_w} Tmol/yr" ,
                        ALK: f"{0} Tmol/yr"},
                  ctype = {DIC: "Regular",
                           ALK: "Regular"},
    )


    """

    def __init__(self, **kwargs) -> None:

        self.__parse_kwargs__(kwargs)

        # # self.source.lor is a  list with the object names in the group
        self.mo = self.sink.lor[0].mo
        self.loc: list = []  # list of connection objects

        if self.mo.debug:
            print(f"mo.register = {self.mo.register}, register ={self.register.name}")
            print(f"name = {self.name}")

        if self.mo.register == "None":  # global name space
            if self.register == "None":
                if self.name == "None":  # set connection group name
                    self.name = f"CG_{self.source.name}2{self.sink.name}{self.id}"

                self.full_name = f"{self.mo.name}.{self.name}"
                if self.mo.debug:
                    print(f"1 fn = {self.full_name}")
            else:  # with registration
                if self.name == "None":
                    self.name = f"CG_{self.source.name}2{self.sink.name}{self.id}"

                self.full_name = f"{self.mo.name}.{self.name}"
                if self.mo.debug:
                    print(f"2 fn = {self.full_name}")
        else:  # local name_space registration
            if self.name == "None":
                self.name = f"CG_{self.source.name}2{self.sink.name}"
                if self.mo.debug:
                    print(f"3 setting name = {self.name}")
            self.full_name = f"{self.mo.name}.{self.name}"
            if self.mo.debug:
                print(f"4 fn = {self.full_name}")

        if self.mo.debug:
            print(f"5 fn = {self.full_name}")

        self.base_name = self.name
        kwargs.update({"name": self.name})  # and add it to the kwargs
        self.__register_name__()

        # register connection group in global namespace
        # m_type="mass_only",
        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        logging.info(f"Created {self.name}")
        self.__create_connections__()

    def update(self, **kwargs) -> None:
        """Add a connection to the connection group
        This will overwrite the original kwargs though

        """

        full_name = self.full_name
        self.__parse_kwargs__(kwargs)
        self.full_name = full_name
        self.__create_connections__()

    def __parse_kwargs__(self, kwargs) -> None:
        """
        Parse and register the keyword arguments

        """
        # provide a dict of all known keywords and their type
        self.lkk: Dict[str, any] = {
            "id": str,
            "name": str,
            "source": (SourceGroup, ReservoirGroup),
            "sink": (SinkGroup, ReservoirGroup),
            "delta": dict,
            "rate": dict,
            "pl": dict,
            "signal": Signal,
            "alpha": dict,
            "species": dict,
            "ctype": dict,
            "ref_reservoirs": dict,
            "ref_flux": dict,
            "plot": dict,
            "scale": dict,
            "bypass": (dict, str),
            "register": any,
            "save_flux_data": (bool, str),
        }

        # list of default values if none provided
        self.lod: Dict[any, any] = {
            "name": "None",
            "id": "",
            "register": "None",
            "save_flux_data": "None",
        }

        # provide a list of absolutely required keywords
        self.lrk: list = ["source", "sink"]
        self.__validateandregister__(kwargs)

        if self.save_flux_data == "None":
            self.save_flux_data = self.register.save_flux_data
            self.kwargs.update({"save_flux_data": self.register.save_flux_data})

    def __create_connections__(self) -> None:
        """Create Connections"""
        # find all sub reservoirs which have been specified by the ctype keyword
        self.connections: list = []
        for r in self.source.lor:
            if self.ctype == "None":
                raise ValueError(
                    f"Connectiongroup {self.name} must specify 'ctype'. See help(Connectiongroup)"
                )
            if r.sp in self.ctype:
                self.connections.append(r)
                logging.debug(f"found species {r.sp.n}")

        # now we need to create defaults for all connections
        # we setup a dict like this  {SO4:{cid:xxx,plot:xxx}}
        self.cd: dict = {}  # connection dictionary
        for r in self.connections:  # ["SO4", "H2S"]
            self.cd[r.n] = {
                "cid": self.id,
                "plot": "yes",
                "delta": "None",
                "alpha": "None",
                "rate": "None",
                "scale": "None",
                "ctype": "None",
                "ref_reservoirs": "None",
                "ref_flux": "None",
                "bypass": "None",
            }

            # print(f"self.cd[r.n] = {self.cd[r.n]}")

            # test defaults against actual keyword value
            for kcd, vcd in self.cd[r.n].items():
                if kcd in self.kwargs:  # found entry like ctype
                    if r.sp in self.kwargs[kcd]:  # {SO4: xxx}
                        # update the entry
                        self.cd[r.n][kcd] = self.kwargs[kcd][r.sp]
            # now we can create the connection

            # print(self.cd)
            # name = f"{r.n}"

            if self.full_name == "None":
                register = self.mo
            else:
                register = self

            a = Connect(
                source=getattr(self.source, r.n),
                sink=getattr(self.sink, r.n),
                rate=self.cd[r.n]["rate"],
                delta=self.cd[r.n]["delta"],
                alpha=self.cd[r.n]["alpha"],
                plot=self.cd[r.n]["plot"],
                ctype=self.cd[r.n]["ctype"],
                scale=self.cd[r.n]["scale"],
                bypass=self.cd[r.n]["bypass"],
                ref_reservoirs=self.cd[r.n]["ref_reservoirs"],
                ref_flux=self.cd[r.n]["ref_flux"],
                save_flux_data=self.save_flux_data,
                groupname=True,
                id=self.id,
                register=register,
            )

            ## add connection to list of connections
            self.loc.append(a)
            if self.mo.debug:
                print(
                    f"created connection with full name {a.full_name}, registered to {self.name} "
                    f"fn = {self.full_name}"
                )

    def info(self) -> None:
        """List all connections in this group"""

        print(f"Group Connection from {self.source.name} to {self.sink.name}\n")
        print("The following Connections are part of this group\n")
        print(f"You can query the details of each connection like this:\n")
        for c in self.loc:
            print(f"{c.name}: {self.name}.{c.name}.info()")

        print("")


class AirSeaExchange(esbmtkBase):
    """The class creates a connection between liquid reservoir (i.e., an
    ocean), and a gas reservoir (i.e., the atmosphere).

    Example :
    ~~~~~~~~

    AirSeaExchange(
        gas_reservoir= must be a gasreservoir
        liquid_reservoir = must be a reservoir
        solubility= as returned by the swc object
        area = Ocean.area, [m^2]
        piston_velocity = 4.8*365, [m/yr]
        id = str, optional
        water_vapor_pressure=Ocean.swc.p_H2O,
        ref_quantity = optional
        )

    In some cases the gas flux does not depend on the main reservoir species
    but on a derived quantity, e.g., [CO2aq]. Specify the ref_quantity
    keyword to point to a different species/calculated species.

    """

    def __init__(self, **kwargs) -> None:
        """initialize instance"""

        from .utility_functions import check_for_quantity

        self.lof: list = []

        self.__check_keywords__(kwargs)

        self.scale = self.area * self.piston_velocity

        # create connection and flux name
        cname = f"C_{self.lr.register.name}_2_{self.gr.name}"
        fname = f"{self.gr.sp.name}_F"

        if self.id == "None" or self.id == "":
            pass
        else:
            fname = f"{fname}_{self.id}"

        # print(f"Using fname = {fname}")
        self.name = cname
        self.full_name = cname

        # initalize a flux instance
        self.fh = Flux(
            name=fname,  # flux name
            species=self.species,  # Species handle
            delta=0,  # delta value of flux
            rate="0 mol/a",  # flux value
            register=self,  # register with this connection
            isotopes=self.isotopes,
        )
        # register flux with liquid reservoir
        self.lr.lof.append(self.fh)
        self.lr.lio[self.fh] = 1  # flux direction

        # register flux with gas reservoir
        self.gr.lof.append(self.fh)
        self.gr.lio[self.fh] = -1  # flux direction
        # register with connection
        self.lof.append(self.fh)

        if self.lr.register == "None":
            swc = self.lr.swc
        else:
            swc = self.lr.register.swc

        # initialize process instance
        ph = GasExchange(
            name="_PGex",
            gas=self.gr,  # gas reservoir
            liquid=self.lr,  # reservoir
            ref_species=self.ref_species,  # concentration
            flux=self.fh,  # flux handle
            register=self.fh,
            scale=self.scale,
            solubility=self.solubility,
            water_vapor_pressure=self.water_vapor_pressure,
            seawaterconstants=swc,
            isotopes=True,
        )

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name__()
        # register process with reservoir
        ph.__register__(self.lr, self.fh)

        # register connector with liquid reservoirgroup
        # spr = getattr(self.lr, self.lr.name)
        self.lr.loc.add(self)
        # register connector with gas reservoir
        self.gr.loc.add(self)
        # register connector with model
        self.mo.loc.add(self)

        logging.info(f"Created {self.full_name}")

    def __check_keywords__(self, kwargs) -> None:
        # provide a dict of all known keywords and their type
        self.lkk: Dict[str, any] = {
            "gas_reservoir": GasReservoir,
            "liquid_reservoir": Reservoir,
            "solubility": float,
            "piston_velocity": (str, Q_),
            "area": float,
            "id": str,
            "name": str,
            "water_vapor_pressure": (Number, np.float64),
            "ref_species": (Reservoir, Number, np.float64, np.ndarray),
            "species": (Species, str),
        }

        # provide a list of absolutely required keywords
        self.lrk: list[str] = [
            "gas_reservoir",
            "liquid_reservoir",
            "solubility",
            "piston_velocity",
            "area",
            "water_vapor_pressure",
            "species",
        ]

        # list of default values if none provided
        self.lod: Dict[str, any] = {"id": "None"}
        self.__initerrormessages__()
        self.bem.update(
            {
                "gas_reservoir": "must be a Reservoir",
                "liquid_reservoir": "must be a Reservoir",
                "solubility": "must be a float number",
                "piston_velocity": "must be a float number",
                "area": "must be a float number",
                "name": "None",
                "ref_species": "None",
            }
        )

        self.__validateandregister__(kwargs)

        # make sure piston velocity is in the right units
        self.piston_velocity = check_for_quantity(self.piston_velocity)
        self.piston_velocity = self.piston_velocity.to(
            f"meter/{self.liquid_reservoir.mo.t_unit}"
        ).magnitude

        # ref_species can point to vr_data fields which are of type
        # numpy array
        if isinstance(self.ref_species, np.ndarray):
            testv = self.ref_species[0]
        else:
            testv = self.ref_species

        if testv == "None":
            self.ref_species = self.species

        self.lr = self.liquid_reservoir
        self.gr = self.gas_reservoir

        if self.species.name in ["CO2", "DIC", "O2"]:
            pass
        else:
            raise ValueError(f"{self.species.name} not implemented yet")

        if isinstance(self.lr.register, ReservoirGroup):
            n = self.lr.register.name
        else:
            n = self.lr.name

        self.name = f"GC_{n}_2_{self.gr.name}_{self.species.name}"

        if self.id == "None" or self.id == "":
            pass
        else:
            self.name = f"{self.name}_{self.id}"

        if self.register == "None":
            self.full_name = self.name
        else:
            self.full_name = f"{self.register.full_name}.{self.name}"

        self.base_name = self.name

        # decide if this connection needs isotope calculations
        if self.gas_reservoir.isotopes:
            self.isotopes = True
        else:
            self.isotopes = False

        self.mo = self.species.mo
        self.model = self.mo



"""esbmtk: A general purpose Earth Science box model toolkit Copyright
     (C), 2020-2021 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or
     modify it under the terms of the GNU General Public License as
     published by the Free Software Foundation, either version 3 of
     the License, or (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
     General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see
     <https://www.gnu.org/licenses/>.

"""

from numbers import Number
from nptyping import *
from typing import *
from numpy import array, set_printoptions, arange, zeros, interp, mean
from copy import deepcopy, copy
from time import process_time
from numba import njit
from numba.typed import List
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

import logging
import time
import builtins
from .esbmtk import esbmtkBase, Model, Reservoir, VirtualReservoir, ReservoirGroup, Flux


# define a transform function to display the Hplus concentration as pH
def phc(m: float) -> float:
    """the reservoir class accepts a plot transform. here we use this to
    display the H+ concentrations as pH. After import, you can use it
    with like this in the reservoir definition

     plot_transform_c=phc,

    """
    import numpy as np

    pH = -np.log10(m)
    return pH


class SeawaterConstants(esbmtkBase):
    """Provide basic seawater properties as a function of T and Salinity.
    Pressure may come at a later stage

    Example:

    Seawater(name="SW",
             model=,
             temperature = optional in C, defaults to 25,
             salinity  = optional in psu, defaults to 35,
             pressure = optional, defaults to 0 bars = 1atm,
             pH = optional, defaults to 8.1,
             units = "mol/l" or "mol/kg"
            )

    useful methods:

    SW.show() will list all known values

    After initialization this class provides access to each value the following way

    instance_name.variable_name

    """

    def __init__(self, **kwargs: Dict[str, str]):

        import math
        from esbmtk import Q_

        # dict of all known keywords and their type
        self.lkk: Dict[str, any] = {
            "name": str,
            "model": Model,
            "salinity": (int, float),
            "temperature": (int, float),
            "pH": (int, float),
            "pressure": Number,
            "register": any,
            "units": (any),
        }

        # provide a list of absolutely required keywords
        self.lrk: list = ["name", "units"]
        # list of default values if none provided
        self.lod: Dict[str, any] = {
            "salinity": 35.0,
            "temperature": 25.0,
            "pH": 8.1,
            "pressure": 0,
            "register": "None",
        }

        # validate input and initialize instance variables
        self.__initerrormessages__()
        self.__validateandregister__(kwargs)  # initialize keyword values

        u1 = Q_("mol/liter").units
        u2 = Q_("mol/kg").units
        if self.units != u1 and self.units != u2:
            raise ValueError(f"units must be {u1} or {u2}")

        # legacy names
        self.n: str = self.name  # string =  name of this instance
        self.mo: Model = self.model
        self.hplus = 10 ** -self.pH
        self.constants: list = ["K0", "K1", "K2", "KW", "KB", "Ksp", "Ksp0", "KS", "KF"]
        self.species: list = [
            "dic",
            "ta",
            "ca",
            "co2",
            "hco3",
            "co3",
            "boron",
            "boh4",
            "boh3",
            "oh",
            "ca2",
            "so4",
            "hplus",
        ]

        self.update()

        if self.mo.register == "local" and self.register == "None":
            self.register = self.mo

        self.__register_name__()

    def update(self, **kwargs: dict) -> None:
        """Update values if necessary"""

        from math import log10
        from esbmtk import Q_

        if kwargs:
            self.lrk: list = []
            self.__validateandregister__(kwargs)

        # update K values and species concentrations according to P, S, and T
        self.__get_density__()
        self.__init_std_seawater__()
        self.__init_bisulfide__()
        self.__init_hydrogen_floride__()
        self.__init_carbon__()
        self.__init_boron__()
        self.__init_water__()
        self.__init_gasexchange__()
        self.__init_calcite__()
        self.__init_c_fractionation_factors__()

        # get total alkalinity
        self.ca = self.hco3 + 2 * self.co3
        self.ta = self.ca + self.boh4 + self.oh - self.hplus

        # convert to mol/liter if necessary
        if self.units == Q_("1 mole/liter").units:
            cf = self.density / 1000  # convert to kg/liter

            # constants and species are just names, so we need to
            # retrieve the actual variable first
            for n in self.constants:
                v = getattr(self, n)
                setattr(self, n, v * cf)
            for n in self.species:
                v = getattr(self, n)
                setattr(self, n, v * cf)
        else:
            "\n Constants are mol/kg! \n"

        # update pk values
        for n in self.constants:
            v = getattr(self, n)
            pk = f"p{n.lower()}"
            setattr(self, pk, -log10(v))

    def show(self) -> None:
        """Printout pK values."""

        from math import log10

        for n in self.species:
            v = getattr(self, n)
            print(f"{n} = {v * 1E6:.2f} nmol/l")

        print(f"pH = {-log10(self.hplus):.2f}\n")
        print(f"salinity = {self.salinity:.2f}")
        print(f"temperature = {self.temperature:.2f}\n")

        for n in self.constants:
            K = getattr(self, n)
            pk = getattr(self, f"p{n.lower()}")
            print(f"{n} = {K:.2e}, p{n} = {pk:.2f}")

    def __get_density__(self):
        """Calculate seawater density as function of temperature,
        pressure and salinity in kg/m^3. Shamelessy copied
        from R. Zeebes equic.m mathlab file.

        TC = temp in C
        P = pressure
        S = salinity
        """

        TC = self.temperature
        P = self.pressure
        S = self.salinity
        # density of pure water
        rhow = (
            999.842594
            + 6.793952e-2 * TC
            - 9.095290e-3 * TC ** 2
            + 1.001685e-4 * TC ** 3
            - 1.120083e-6 * TC ** 4
            + 6.536332e-9 * TC ** 5
        )

        # density of of seawater at 1 atm, P=0
        A = (
            8.24493e-1
            - 4.0899e-3 * TC
            + 7.6438e-5 * TC ** 2
            - 8.2467e-7 * TC ** 3
            + 5.3875e-9 * TC ** 4
        )
        B = -5.72466e-3 + 1.0227e-4 * TC - 1.6546e-6 * TC ** 2
        C = 4.8314e-4
        rho0 = rhow + A * S + B * S ** (3 / 2) + C * S ** 2

        """Secant bulk modulus of pure water is the average change in
        pressure divided by the total change in volume per unit of
        initial volume.
        """
        Ksbmw = (
            19652.21
            + 148.4206 * TC
            - 2.327105 * TC ** 2
            + 1.360477e-2 * TC ** 3
            - 5.155288e-5 * TC ** 4
        )
        # Secant bulk modulus of seawater at 1 atm
        Ksbm0 = (
            Ksbmw
            + S * (54.6746 - 0.603459 * TC + 1.09987e-2 * TC ** 2 - 6.1670e-5 * TC ** 3)
            + S ** (3 / 2) * (7.944e-2 + 1.6483e-2 * TC - 5.3009e-4 * TC ** 2)
        )
        # Secant modulus of seawater at S,T,P
        Ksbm = (
            Ksbm0
            + P
            * (3.239908 + 1.43713e-3 * TC + 1.16092e-4 * TC ** 2 - 5.77905e-7 * TC ** 3)
            + P * S * (2.2838e-3 - 1.0981e-5 * TC - 1.6078e-6 * TC ** 2)
            + P * S ** (3 / 2) * 1.91075e-4
            + P * P * (8.50935e-5 - 6.12293e-6 * TC + 5.2787e-8 * TC ** 2)
            + P ** 2 * S * (-9.9348e-7 + 2.0816e-8 * TC + 9.1697e-10 * TC ** 2)
        )
        # Density of seawater at S,T,P in kg/m^3
        self.density = rho0 / (1.0 - P / Ksbm)

    def __init_std_seawater__(self) -> None:
        """Provide values for standard seawater. Data after Zeebe and Gladrow
        all values in mol/kg. All values after Zeebe and Gladrow 2001

        """

        self.dic = 0.00204
        self.boron = 0.00042
        self.oh = 0.00001
        self.so4 = 2.7123 / 96
        self.ca2 = 0.01028
        self.Ksp0 = 4.29e-07  # after after Boudreau et al 2010

    def __init_hydrogen_floride__(self) -> None:
        """Bisulfide ion concentration after Dickson 1994, cf.
        Zeebe and Gladrow 2001, p 260

        """

        import numpy as np

        T = 273.15 + self.temperature
        S = self.salinity
        I = (19.924 * S) / (1000 - 1.005 * S)

        lnKF = (
            1590.2 / T
            - 12.641
            + 1.525 * I ** 0.5
            + np.log(1 - 0.001005 * S)
            + np.log(1 + self.ST / self.KS)
        )

        self.KF = np.exp(lnKF)
        self.FT = 7e-5 * self.salinity / 35

    def __init_bisulfide__(self) -> None:
        """Bisulfide ion concentration after Dickson 1994, cf.
        Zeebe and Gladrow 2001, p 260

        """

        import numpy as np

        T = 273.15 + self.temperature
        S = self.salinity
        I = (19.924 * S) / (1000 - 1.005 * S)
        lnKS = (
            -4276.1 / T
            + 141.328
            - 23.093 * np.log(T)
            + (-13856 / T + 324.57 - 47.986 * np.log(T)) * I ** 0.5
            + (35474 / T - 771.54 + 114.723 * np.log(T)) * I
            - 2698 / T * I ** 1.5
            + 1776 / T * I ** 2
            + np.log(1 - 0.001005 * S)
        )

        self.KS = np.exp(lnKS)
        self.ST = self.so4 * self.salinity / 35

    def __init_gasexchange__(self) -> None:
        """Initialize constants for gas-exchange processes"""

        self.water_vapor_partial_pressure()
        self.co2_solubility_constant()
        self.o2_solubility_constant()

    def __init_carbon__(self) -> None:
        """Calculate the carbon equilibrium values as function of
        temperature T and salinity S

        """

        from math import exp, log, log10

        T = 273.15 + self.temperature
        S = self.salinity

        # After Weiss 1974
        lnK0: float = (
            93.4517 * 100 / T
            - 60.2409
            + 23.3585 * log(T / 100)
            + S * (0.023517 - 0.023656 * T / 100 + 0.0047036 * (T / 100) ** 2)
        )

        lnk1: float = (
            -2307.1266 / T
            + 2.83655
            - 1.5529413 * log(T)
            + S ** 0.5 * (-4.0484 / T - 0.20760841)
            + S * 0.08468345
            + S ** (3 / 2) * -0.00654208
            + log(1 - 0.001006 * S)
        )

        lnk2: float = (
            -9.226508
            - 3351.6106 / T
            - 0.2005743 * log(T)
            + (-0.106901773 - 23.9722 / T) * S ** 0.5
            + 0.1130822 * S
            - 0.00846934 * S ** 1.5
            + log(1 - 0.001006 * S)
        )

        self.K0: float = exp(lnK0)
        self.K1: float = exp(lnk1)
        self.K2: float = exp(lnk2)

        self.K1 = self.__pressure_correction__("K1", self.K1)
        self.K2 = self.__pressure_correction__("K2", self.K2)

        self.co2 = self.dic / (
            1 + self.K1 / self.hplus + self.K1 * self.K2 / self.hplus ** 2
        )
        self.hco3 = self.dic / (1 + self.hplus / self.K1 + self.K2 / self.hplus)
        self.co3 = self.dic / (
            1 + self.hplus / self.K2 + self.hplus ** 2 / (self.K1 * self.K2)
        )

    def __init_boron__(self) -> None:
        """Calculate the boron equilibrium values as function of
        temperature T and salinity S

        """

        from math import exp, log

        T = 273.15 + self.temperature
        S = self.salinity

        lnkb = (
            (
                -8966.9
                - 2890.53 * S ** 0.5
                - 77.942 * S
                + 1.728 * S ** 1.5
                - 0.0996 * S ** 2
            )
            / T
            + 148.0248
            + 137.1942 * S ** 0.5
            + 1.62142 * S
            - (24.4344 + 25.085 * S ** 0.5 + 0.2474 * S) * log(T)
            + 0.053105 * S ** 0.5 * T
        )

        self.KB = exp(lnkb)
        self.KB = self.__pressure_correction__("KB", self.KB)

        self.boh4 = self.boron * self.KB / (self.hplus + self.KB)
        self.boh3 = self.boron - self.boh4

    def __init_water__(self) -> None:
        """Calculate the water equilibrium values as function of
        temperature T and salinity S

        """

        from math import exp, log

        T = 273.15 + self.temperature
        S = self.salinity

        lnKW = (
            148.96502
            - 13847.27 / T
            - 23.6521 * log(T)
            + (118.67 / T - 5.977 + 1.0495 * log(T)) * S ** 0.5
            - 0.01615 * S
        )
        self.KW = exp(lnKW)
        self.KW = self.__pressure_correction__("KW", self.KW)
        self.oh = self.KW / self.hplus

    def __pressure_correction__(self, n: str, K: float) -> float:
        """Correct K-values for pressure. After Zeebe and Wolf Gladrow 2001

        name = name of K-value, i.e. "K1"
        K = uncorrected value
        T = temperature in Deg C
        P = pressure in atm
        """

        from math import exp, log

        R: float = 83.131
        Tc: float = self.temperature
        T: float = 273.15 + Tc
        P: float = self.pressure
        RT: float = R * T

        A: dict = {}
        A["K1"]: list = [25.50, 0.1271, 0.0, 3.08, 0.0877]
        A["K2"]: list = [15.82, -0.0219, 0.0, -1.13, -0.1475]
        A["KB"]: list = [29.48, 0.1622, -2.6080, 2.84, 0.0]
        A["KW"]: list = [25.60, 0.2324, -3.6246, 5.13, 0.0794]
        A["KS"]: list = [18.03, 0.0466, 0.3160, 4.53, 0.0900]
        A["KF"]: list = [9.780, -0.0090, -0.942, 3.91, 0.054]
        A["Kca"]: list = [48.76, 0.5304, 0.0, 11.76, 0.3692]
        A["Kar"]: list = [46.00, 0.5304, 0.0, 11.76, 0.3692]

        a: list = A[n]

        DV: float = -a[0] + (a[1] * Tc) + (a[2] / 1000 * Tc ** 2)
        DK: float = -a[3] / 1000 + (a[4] / 1000 * Tc) + (0 * Tc ** 2)

        # print(f"DV = {DV}")
        # print(f"DK = {DK}")
        # print(f"log k= {log(K)}")

        lnkp: float = -(DV / RT) * P + (0.5 * DK / RT) * P ** 2 + log(K)
        # print(lnkp)

        return exp(lnkp)

    def water_vapor_partial_pressure(self) -> None:
        """Calculate the water vapor partial pressure at sealevel (1 atm) as
        a function of temperature and salinity. Eq. Weiss and Price 1980
        doi:10.1016/0304-4203(80)90024-9

        Since we assume that we only use this expression at sealevel,
        we drop the pressure term

        The result is in p/1atm (i.e., a percentage)
        """

        T = self.temperature + 273.15
        S = self.salinity

        self.p_H2O = np.exp(
            24.4543 - 67.4509 * (100 / T) - 4.8489 * np.log(T / 100) - 0.000544 * S
        )

    def co2_solubility_constant(self) -> None:
        """Calculate the solubility of CO2 at a given temperature and salinity.
        Coefficients after Sarmiento and Gruber 2006 which includes
        corrections for CO2 to correct for non ideal gas behavior

        Parameters Ai & Bi from Tab 3.2.2 in  Sarmiento and Gruber 2006

        The result is in mol/(m^3 * atm)
        """

        # Calculate the volumetric solubility function F_A in mol/l
        S = self.salinity  # unitless
        T = 273.15 + self.temperature  # C
        A1 = -160.7333
        A2 = 215.4152
        A3 = 89.892
        A4 = -1.47759
        B1 = 0.029941
        B2 = -0.027455
        B3 = 0.0053407

        # F in mol/(l * atm)
        F = self.calc_solubility_term(S, T, A1, A2, A3, A4, B1, B2, B3)

        # correct for water vapor partial pressure
        self.SA_co2 = F / (1 - self.p_H2O)  # mol/(m^3 * atm)

    def o2_solubility_constant(self) -> None:
        """Calculate the solubility of CO2 at a given temperature and salinity. Coefficients
        after Sarmiento and Gruber 2006 which includes corrections for CO2 to correct for non
        ideal gas behavior

        Parameters Ai & Bi from Tab 3.2.2 in  Sarmiento and Gruber 2006

        The result is in mol/(m^3 atm)
        """

        # Calculate the volumetric solubility function F_A in mol/l/m^3
        S = self.salinity  # unit less
        T = 273.15 + self.temperature  # in C
        A1 = -58.3877
        A2 = 85.8079
        A3 = 23.8439
        A4 = 0
        B1 = -0.034892
        B2 = 0.015568
        B3 = -0.0019387

        b = self.calc_solubility_term(S, T, A1, A2, A3, A4, B1, B2, B3)

        # and convert from bunsen coefficient to solubility
        VA = 22.4136  # after Sarmiento & Gruber 2006
        self.SA_o2 = b / VA

    def calc_solubility_term(self, S, T, A1, A2, A3, A4, B1, B2, B3) -> float:
        ln_F = (
            A1
            + A2 * (100 / T)
            + A3 * np.log(T / 100)
            + A4 * (T / 100) ** 2
            + S * (B1 + B2 * (T / 100) + B3 * (T / 100) ** 2)
        )
        F = np.exp(ln_F) * 1000  # to get mol/(m^3 atm)

        return F

    def __init_calcite__(self) -> None:
        """Calculate Calcite solubility as a function of pressure following
        Fig 1 in in Boudreau et al, 2010, https://doi.org/10.1029/2009gl041847

        Note that this equation assumes an idealized ocean temperature profile.
        So it cannot be applied to a warm ocean

        """

        self.Ksp = 4.3513e-7 * np.exp(0.0019585 * self.pressure)

    def __init_c_fractionation_factors__(self):
        """Calculate the fractionation factors for the various carbon species transitions.
        After Zeebe and Gladrow, 2001, Chapter 3.2.3, page 186

        e = (a -1) * 1E3

        and

        a =  1 + e / 1E3

        where the subscripts denote:

        g = gaseous CO2
        d = dissolved CO2
        b = bicarbonate ion
        c = carbonate ion

        """

        T = 273.15 + self.temperature

        # CO2g versus HCO3, e = epsilon, a = alpha
        self.e_gb: float = -9483 / T + 23.89
        self.a_gb: float = 1 + self.e_gb / 1000

        # CO2aq versus CO2g
        self.e_dg: float = -373 / T + 0.19
        self.a_dg: float = 1 + self.e_dg / 1000

        # CO2aq versus HCO3
        self.e_db: float = -9866 / T + 24.12
        self.a_db: float = 1 + self.e_db / 1000

        # CO32- versus HCO3
        self.e_cb: float = -867 / T + 2.52
        self.a_cb: float = 1 + self.e_cb / 1000

        # kinetic fractionation during gas exchange
        # parameters after Zhang et. al.1995
        # https://doi.org/10.1016/0016-7037(95)91550-D
        m = 0.14 / 16
        c = m * 5 + 0.95
        self.e_u: float = self.temperature * m - c
        self.a_u: float = 1 + self.e_u / 1000


def calc_pCO2(
    dic: Union[Reservoir, VirtualReservoir],
    hplus: Union[Reservoir, VirtualReservoir],
    SW: SeawaterConstants,
) -> Union[NDArray, Float]:

    """
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic: Reservoir  = DIC concentrations in mol/liter
    hplus: Reservoir = H+ concentrations in mol/liter
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2(dic,h,SW)

     DataField(name = "SurfaceWaterpCO2",
                       associated_with = reservoir_handle,
                       y1_data = pco2,
                       y1_label = r"pCO_{2}",
                       y1_legend = r"pCO_{2}",
                       )
    Author: T. Tsan

    """

    dic_c: [NDArray, Float] = dic.c
    hplus_c: [NDArray, Float] = hplus.c

    k1: float = SW.K1
    k2: float = SW.K2

    co2: [NDArray, Float] = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c ** 2)))

    pco2: [NDArray, Float] = co2 / SW.K0 * 1e6

    return pco2


def calc_pCO2b(
    dic: Union[float, NDArray],
    hplus: Union[float, NDArray],
    SW: SeawaterConstants,
) -> Union[NDArray, Float]:

    """
    Same as calc_pCO2, but accepts values/arrays rather than Reservoirs.
    Calculate the concentration of pCO2 as a function of DIC,
    H+, K1 and k2 and returns a numpy array containing
    the pCO2 in uatm at each timestep. Calculations are based off
    equations from Follows, 2006. doi:10.1016/j.ocemod.2005.05.004
    dic:  = DIC concentrations in mol/liter
    hplus: = H+ concentrations in mol/liter
    SW: Seawater = Seawater object for the model
    it is typically used with a DataField object, e.g.
    pco2 = calc_pCO2b(dic,h,SW)
     DataField(name = "SurfaceWaterpCO2",
                       associated_with = reservoir_handle,
                       y1_data = pco2b,
                       y1_label = r"pCO_{2}",
                       y1_legend = r"pCO_{2}",
                       )
    """

    dic_c: [NDArray, Float] = dic

    hplus_c: [NDArray, Float] = hplus

    k1: float = SW.K1
    k2: float = SW.K2

    co2: [NDArray, Float] = dic_c / (1 + (k1 / hplus_c) + (k1 * k2 / (hplus_c ** 2)))

    pco2: [NDArray, Float] = co2 / SW.K0 * 1e6

    return pco2


@njit(parallel=False, fastmath=True, error_model="numpy")
def calc_carbonates_1(i: int, input_data: List, vr_data: List, params: List) -> None:
    """Calculates and returns the carbonate concentrations and saturation state
     at the ith time-step of the model.

    The function assumes that vr_data will be in the following order:
        [H+, CA, HCO3, CO3, CO2(aq), omega]

    LIMITATIONS:
    - This in used in conjunction with ExternalCode objects!
    - Assumes all concentrations are in mol/L
    - Assumes your Model is in mol/L ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    See add_carbonate_system_1 in utility_functions.py on how to call this function

    Author: M. Niazi & T. Tsan, 2021
    """

    k1 = params[0]  # K1
    k2 = params[1]  # K2
    KW = params[2]  # KW
    KB = params[3]  # KB
    boron = params[4]  # boron
    ca2 = params[5]  # Ca2+
    ksp = params[6]  # Ksp

    dic: float = input_data[0][i - 1]
    ta: float = input_data[1][i - 1]
    hplus: float = vr_data[0][i - 1]

    # calculates carbonate alkalinity (ca) based on H+ concentration from the
    # previous time-step
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # hplus
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1 * k1 - 4 * k1 * k2 * (1 - (2 * gamm))

    hplus: float = 0.5 * ((gamm - 1) * k1 + (dummy ** 0.5))
    # hco3 and co3
    """ Since CA = [hco3] + 2[co3], can the below expression can be simplified
    """
    # co3: float = dic / (1 + (hplus / k2) + ((hplus ** 2) / (k1 * k2)))
    hco3: float = dic / (1 + (hplus / k1) + (k2 / hplus))
    co3: float = (ca - hco3) / 2
    # co2 (aq)
    """DIC = hco3 + co3 + co2 + H2CO3 The last term is however rather
    small, so it may be ok to simply write co2aq = dic - hco3 + co3.
    Let's test this once we have a case where pco2 is calculated from co2aq
    """
    #  co2aq: float = dic / (1 + (k1 / hplus) + (k1 * k2 / (hplus ** 2)))
    co2aq: float = dic - hco3 - co3
    # omega: float = ca2 * co3 / ksp

    vr_data[0][i] = hplus
    vr_data[1][i] = ca
    vr_data[2][i] = hco3
    vr_data[3][i] = co3
    vr_data[4][i] = co2aq
    # vr_data[5][i] = omega


@njit(fastmath=True, error_model="numpy")
def calc_carbonates_2(i: int, input_data: List, vr_data: List, params: List) -> None:
    """Calculates and returns the carbonate concentrations and carbonate compensation
    depth (zcc) at the ith time-step of the model.

    The function assumes that vr_data will be in the following order:
        [H+, CA, HCO3, CO3, CO2(aq), zsat, zcc, zsnow, Fburial,
        B, BNS, BDS_under, BDS_resp, BDS, BCC, BPDC, BD,omega]

    LIMITATIONS:
    - This in used in conjunction with ExternalCode objects!
    - Assumes all concentrations are in mol/L
    - Assumes your Model is in mol/L ! Otherwise, DIC and TA updating will not
    be correct.

    Calculations are based off equations from:
    Boudreau et al., 2010, https://doi.org/10.1029/2009GB003654
    Follows, 2006, doi:10.1016/j.ocemod.2005.05.004

    See add_carbonate_system_2 in utility_functions.py on how to call this function.
    The input data is a follows

        reservoir DIC.m,  # 0
        reservoir DIC.l,  # 1
        reservoir DIC.c,  # 2
        reservoir TA.m,  # 3 TA mass
        reservoir.TA.c,  # 4 TA concentration
        Export_flux.fa,  # 5
        area_table,  # 6
        area_dz_table,  # 7
        Csat_table,  # 8
        reservoir.DIC.v,  # 9 reservoir volume

    Author: M. Niazi & T. Tsan, 2021
    """

    # Parameters
    k1 = params[0]
    k2 = params[1]
    KW = params[2]
    KB = params[3]
    boron = params[4]
    ksp0 = params[5]
    kc = params[6]
    volume = params[7]
    AD = params[8]
    zsat0 = int(abs(params[9]))
    ca2 = params[10]
    dt = params[11]
    I_caco3 = params[12]
    alpha = params[13]
    zsat_min = int(abs(params[14]))
    zmax = int(abs(params[15]))
    z0 = int(abs(params[16]))
    ksp = params[17]

    # Data
    dic: float = input_data[2][i - 1]  # DIC concentration [mol/l]
    ta: float = input_data[4][i - 1]  # TA concentration [mol/l]
    Bm: float = input_data[5][0]  # Carbonate Export Flux [mol/yr]
    B12: float = input_data[5][1]  # Carbonate Export Flux light isotope
    v: float = input_data[9][i - 1]  # volume
    # lookup tables
    depth_area_table: NDArray = input_data[6]  # depth look-up table
    area_dz_table: NDArray = input_data[7]  # area_dz table
    Csat_table: NDArray = input_data[8]  # Csat table

    # vr_data
    hplus: float = vr_data[0][i - 1]  # H+ concentration [mol/l]
    zsnow = vr_data[7][i - 1]  # previous zsnow

    # calc carbonate alkalinity based t-1
    oh: float = KW / hplus
    boh4: float = boron * KB / (hplus + KB)
    fg: float = hplus - oh - boh4
    ca: float = ta + fg

    # calculate carbon speciation
    # The following equations are after Follows et al. 2006
    gamm: float = dic / ca
    dummy: float = (1 - gamm) * (1 - gamm) * k1 * k1 - 4 * k1 * k2 * (1 - (2 * gamm))

    hplus: float = 0.5 * ((gamm - 1) * k1 + (dummy ** 0.5))
    # co3: float = dic / (1 + (hplus / k2) + ((hplus ** 2) / (k1 * k2)))
    hco3: float = dic / (1 + (hplus / k1) + (k2 / hplus))
    co3: float = (ca - hco3) / 2
    # DIC = hco3 + co3 + co2 + H2CO3 The last term is however rather
    # small, so it may be ok to simply write co2aq = dic - hco3 + co3.
    co2aq: float = dic - co3 - hco3
    # co2aq: float = dic / (1 + (k1 / hplus) + (k1 * k2 / (hplus ** 2)))
    # omega: float = (ca2 * co3) / ksp

    # ---------- compute critical depth intervals eq after  Boudreau (2010)
    # all depths will be positive to facilitate the use of lookup_tables

    # prevent co3 from becoming zero
    if co3 <= 0:
        co3 = 1e-16

    zsat = int(max((zsat0 * np.log(ca2 * co3 / ksp0)), zsat_min))  # eq2
    if zsat < zsat_min:
        zsat = int(zsat_min)

    zcc = int(zsat0 * np.log(Bm * ca2 / (ksp0 * AD * kc) + ca2 * co3 / ksp0))  # eq3

    # ---- Get fractional areas
    B_AD = Bm / AD

    if zcc > zmax:
        zcc = int(zmax)
    if zcc < zsat_min:
        zcc = zsat_min

    A_z0_zsat = depth_area_table[z0] - depth_area_table[zsat]
    A_zsat_zcc = depth_area_table[zsat] - depth_area_table[zcc]
    A_zcc_zmax = depth_area_table[zcc] - depth_area_table[zmax]

    # ------------------------Calculate Burial Fluxes------------------------------------
    # BCC = (A(zcc, zmax) / AD) * B, eq 7
    BCC = A_zcc_zmax * B_AD

    # BNS = alpha_RD * ((A(z0, zsat) * B) / AD) eq 8
    BNS = alpha * A_z0_zsat * B_AD

    # BDS_under = kc int(zcc,zsat) area' Csat(z,t) - [CO3](t) dz, eq 9a
    diff_co3 = Csat_table[zsat:zcc] - co3
    area_p = area_dz_table[zsat:zcc]

    BDS_under = kc * area_p.dot(diff_co3)
    BDS_resp = alpha * (A_zsat_zcc * B_AD - BDS_under)
    BDS = BDS_under + BDS_resp

    # BPDC =  kc int(zsnow,zcc) area' Csat(z,t) - [CO3](t) dz, eq 10
    if zcc < zsnow:  # zcc cannot
        if zsnow > zmax:  # zsnow cannot exceed ocean depth
            zsnow = zmax

        diff = Csat_table[zcc : int(zsnow)] - co3
        area_p = area_dz_table[zcc : int(zsnow)]
        BPDC = kc * area_p.dot(diff)
        # eq 4 dzsnow/dt = Bpdc(t) / (a'(zsnow(t)) * ICaCO3
        zsnow = zsnow - BPDC / (area_dz_table[int(zsnow)] * I_caco3) * dt

    else:  # zcc > zsnow
        # there is no carbonate below zsnow, so BPDC = 0
        zsnow = zcc
        BPDC = 0

    # BD & F_burial
    BD: float = BDS + BCC + BNS + BPDC
    Fburial = Bm - BD
    Fburial12 = Fburial * input_data[1][i-1] / input_data[0][i-1]
    diss =  (Bm - Fburial) * dt # dissolution flux 
    diss12 =  (B12 - Fburial12) * dt #  # dissolution flux light isotope

    # # print("{Fburial}.format(")
    # print(Bm)
    # print(Fburial)
    # print(diss)
    # print()
    # # print('df ={:.2e}\n'.format(diss/dt))

    """ Now that the fluxes are known we need to update the reservoirs.
    The concentration in the in the DIC (and TA) of this box are
    DIC.m[i] + Export Flux - Burial Flux, where the isotope ratio
    the Export flux is determined by the overlying box, and the isotope ratio
    of the burial flux is determined by the isotope ratio of this box
    """

    # Update DIC in the deep box
    input_data[0][i] = input_data[0][i] + diss  # DIC
    input_data[1][i] = input_data[1][i] + diss12  # 12C
    input_data[2][i] = input_data[0][i] / v  # DIC concentration

    # Update TA in deep box
    input_data[3][i] = input_data[3][i] + 2 * diss  # TA
    input_data[4][i] = input_data[3][i] / v  # TA concentration

    # copy results into datafields
    vr_data[0][i] = hplus  # 0
    vr_data[1][i] = ca  # 1
    vr_data[2][i] = hco3  # 2
    vr_data[3][i] = co3  # 3
    vr_data[4][i] = co2aq  # 4
    vr_data[5][i] = zsat  # 5
    vr_data[6][i] = zcc  # 6
    vr_data[7][i] = zsnow  # 7
    vr_data[8][i] = Fburial  # 8
    vr_data[9][i] = Fburial12  # 9
    vr_data[10][i] = diss/dt  # 9
    vr_data[11][i] = Bm  # 9



from __future__ import annotations

from pint import UnitRegistry

ureg = UnitRegistry(on_redefinition="ignore")
Q_ = ureg.Quantity

ureg.define("Sverdrup = 1e6 * meter **3 / second = Sv = Sverdrups")
ureg.define("Mol = 1 * mol / liter = M")
ureg.define("fraction = [] = frac")
ureg.define("percent = 1e-2 frac = pct")
ureg.define("permil = 1e-3 fraction")
ureg.define("ppm = 1e-6 fraction")

import numpy as np
np.seterr(invalid='ignore')

from .esbmtk import *
from .extended_classes import *
from .connections import ConnectionGroup, Connection, Connect, AirSeaExchange
from .utility_functions import *
from .sealevel import *
from .solver import *



"""
     esbmtk: A general purpose Earth Science box model toolkit
     Copyright (C), 2020 Ulrich G. Wortmann

     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation, either version 3 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

# from pint import UnitRegistry
from nptyping import *
from typing import *
from numpy import array, set_printoptions, arange, zeros, interp, mean
import time
from time import process_time
import numba
from numba.core import types
from numba import njit, prange
from numba.typed import List

import numpy as np

# import pandas as pd
# import mpmath

import time
import math


def get_l_mass(m: float, d: float, r: float) -> float:
    """
    Calculate the light isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio
    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    return l


def get_imass(m: float, d: float, r: float) -> [float, float]:
    """
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio
    species

    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    h: float = m - l
    return [l, h]


# @njit()
def get_frac(m: float, l: float, a: float) -> [float, float]:
    """Calculate the effect of the istope fractionation factor alpha on
    the ratio between the light and heavy isotope.

    Note that alpha needs to be given as fractional value, i.e., 1.07 rather
    than 70 (i.e., (alpha-1) * 1000

    """

    if a > 1.1 or a < 0.9:
        raise ValueError("alpha needs to be given as fractional value not in permil")

    li: float = -l * m / (a * l - a * m - l)
    hi: float = m - li  # get the new heavy isotope value
    return li, hi


# @njit()
def get_flux_data(m: float, d: float, r: float) -> [NDArray, float]:
    """
    Calculate the isotope masses from bulk mass and delta value.
    Arguments are m = mass, d= delta value, r = abundance ratio
    species. Unlike get_mass, this function returns the full array

    """

    l: float = (1000.0 * m) / ((d + 1000.0) * r + 1000.0)
    h: float = m - l

    return np.array([m, l, h, d])


# @njit()
def get_delta(
    l: [NDArray, [Float64]], h: [NDArray, [Float64]], r: float
) -> [NDArray, [Float64]]:
    """Calculate the delta from the mass of light and heavy isotope
    Arguments are l and h which are the masses of the light and
    heavy isotopes respectively, r = abundance ratio of the
    respective element. Note that this equation can result in a
    siginificant loss of precision (on the order of 1E-13). I
    therefore round the results to numbers largers 1E12 (otherwise a
    delta of zero may no longer be zero)

    """

    return 1000 * (h / l - r) / r


def get_delta_i(l: float, h: float, r: float) -> float:
    """Calculate the delta from the mass of light and heavy isotope
    Arguments are l and h which are the masses of the light and
    heavy isotopes respectively, r = abundance ratio of the
    respective element. Note that this equation can result in a
    siginificant loss of precision (on the order of 1E-13). I
    therefore round the results to numbers largers 1E12 (otherwise a
    delta of zero may no longer be zero)

    """
    # d = 1000 * (h / l - r) / r
    if l > 0:
        d: float = 1e3 * (h / l - r) / r
    else:
        d = 0
    return d


def get_flux_delta(f) -> float:
    """Calculate the delta of flux f"""

    m = f.fa[0]
    l = f.fa[1]
    h = m - l
    r = f.species.r
    d = 1e3 * (h / l - r) / r
    return d


def get_delta_h(h) -> float:
    """Calculate the delta of a flux or reserevoir from total mass
    and mass of light isotope.

    f = flux or reserevoir handle

    returns d a vector of delta values

    """

    r = h.species.r

    d = np.where(h.l > 0, 1e3 * ((h.m - h.l) / h.l - r) / r, 0)

    return d


def execute(
    new: [NDArray, Float64],
    time: [NDArray, Float64],
    lop: list,
    lor: list,
    lpc_f: list,
    lpc_r: list,
) -> None:
    """This is the original object oriented solver"""

    start: float = process_time()
    i = 1  # some processes refer to the previous time step
    for t in time[1:-1]:  # loop over the time vector except the first
        # we first need to calculate all fluxes
        # for r in lor:  # loop over all reservoirs
        #     for p in r.lop:  # loop over reservoir processes
        #         p(i)  # update fluxes

        for p in lop:  # loop over reservoir processes
            p(i)  # update fluxes

        for p in lpc_f:  # update all process based fluxes.
            p(i)

        for r in lor:  # loop over all reservoirs
            flux_list = r.lof

            m = l = 0
            for f in flux_list:  # do sum of fluxes in this reservoir
                direction = r.lio[f]
                m += f.fa[0] * direction  # current flux and direction
                l += f.fa[1] * direction  # current flux and direction

            r.m[i] = r.m[i - 1] + m * r.mo.dt
            r.l[i] = r.l[i - 1] + l * r.mo.dt
            r.c[i] = r.m[i] / r.v[i]

        # update reservoirs which do not depend on fluxes but on
        # functions
        for p in lpc_r:
            p(i)

        i = i + 1

    duration: float = process_time() - start
    print(f"\n Execution time {duration:.2e} cpu seconds\n")


def execute_e(model, new, lop, lor, lpc_f, lpc_r):
    """ """

    # numba.set_num_threads(2)

    # this has nothing todo with self.time below!

    dt: float = lor[0].mo.dt

    if model.first_start:
        start: float = process_time()
        (
            model.fn_vr,
            model.input_data,
            model.vr_data,
            model.vr_params,
            model.count,
        ) = build_vr_list(lpc_r)

        model.fn, model.da, model.pc = build_process_list(lor, lop)
        model.a, model.b, model.c, model.d, model.e = build_flux_lists_all(lor)
        """
        fn = typed list of functions (processes)
        da = process data
        pc = process parameters
        a = reservoir_list
        b = flux list
        c = direction list
        d = virtual reserevoir list
        e = r0 list ???
        """
        model.first_start = False

        duration: float = process_time() - start
        print(f"\n Setup time {duration} cpu seconds\n")

    print("Starting solver")

    wts = time.time()
    start: float = process_time()

    if model.count > 0:
        if len(model.lop) > 0:
            foo(
                model.fn_vr,
                model.input_data,
                model.vr_data,
                model.vr_params,
                model.fn,
                model.da,
                model.pc,
                model.a,  # reserevoir list
                model.b,  # flux list
                model.c,  # direction list
                model.d,  # r0 list
                model.e,
                model.time[:-1],
                model.dt,
            )
        else:
            foo_no_p(
                model.fn_vr,
                model.input_data,
                model.vr_data,
                model.vr_params,
                model.fn,
                model.a,  # reserevoir list
                model.b,  # flux list
                model.c,  # direction list
                model.d,  # r0 list
                model.e,
                model.time[:-1],
                model.dt,
            )
    else:
        foo_no_vr( 
            model.fn,
            model.da,
            model.pc,
            model.a,
            model.b,
            model.c,
            model.d,
            model.e,
            model.time[:-1],
            model.dt,
        )

    duration: float = process_time() - start
    wcd = time.time() - wts
    print(f"\n Total solver time {duration} cpu seconds, wt = {wcd}\n")


# from numba import jit
@njit(parallel=False, fastmath=True, error_model="numpy")
def foo(fn_vr, input_data, vr_data, vr_params, fn, da, pc, a, b, c, d, e, maxt, dt):
    """
    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reservoir list
    e = r0 list ???
    """

    i = 1
    for t in maxt:

        # loop over function (process) list and compute fluxes
        j = 0
        for f in enumerate(fn):
            fn[j](da[j], pc[j], i)
            j = j + 1

        # calculate the resulting reservoir concentrations
        # summarize_fluxes(a, b, c, d, e, i, dt)
        r_steps: int = len(b)
        # loop over reservoirs
        for j in range(r_steps):
            # for j, r in enumerate(b):  # this will catch the list for each reservoir

            # sum fluxes in each reservoir
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        # # update reservoirs which do not depend on fluxes but on
        # # functions
        for j, f in enumerate(fn_vr):
            fn_vr[j](i, input_data[j], vr_data[j], vr_params[j])

        i = i + 1  # next time step


@njit(parallel=False, fastmath=True, error_model="numpy")
def foo_no_p(fn_vr, input_data, vr_data, vr_params, fn, a, b, c, d, e, maxt, dt):
    """
    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reservoir list
    e = r0 list ???
    """

    i = 1
    for t in maxt:

        # calculate the resulting reservoir concentrations
        # summarize_fluxes(a, b, c, d, e, i, dt)
        r_steps: int = len(b)
        # loop over reservoirs
        for j in range(r_steps):
            # for j, r in enumerate(b):  # this will catch the list for each reservoir

            # sum fluxes in each reservoir
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        # # update reservoirs which do not depend on fluxes but on
        # # functions
        for j, f in enumerate(fn_vr):
            fn_vr[j](i, input_data[j], vr_data[j], vr_params[j])

        i = i + 1  # next time step


@njit(parallel=False, fastmath=True, error_model="numpy")
def foo_no_vr(fn, da, pc, a, b, c, d, e, maxt, dt):
    """Same as foo but no virtual reservoirs present

    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reserevoir list
    e = r0 list ???
    """

    i = 1
    for t in maxt:

        # loop over processes
        j = 0
        for f in enumerate(fn):
            fn[j](da[j], pc[j], i)
            j = j + 1

        # calculate the resulting reservoir concentrations
        # summarize_fluxes(a, b, c, d, e, i, dt)
        r_steps: int = len(b)
        # loop over reservoirs
        for j in range(r_steps):
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        i = i + 1


@njit(parallel=False, fastmath=True, error_model="numpy")
def foo_no_vr_no_p(fn, a, b, c, d, e, maxt, dt):
    """Same as foo but no virtual reservoirs present

    fn = flux process list
    fn = typed list of functions (processes)
    da = process data
    pc = process parameters
    a = reservoir_list
    b = flux list
    c = direction list
    d = virtual reserevoir list
    e = r0 list ???
    """

    i = 1
    for t in maxt:

        # calculate the resulting reservoir concentrations
        # summarize_fluxes(a, b, c, d, e, i, dt)
        r_steps: int = len(b)
        # loop over reservoirs
        for j in range(r_steps):
            mass = li = 0.0
            f_steps = len(b[j])
            for u in range(f_steps):
                direction = c[j][u]
                mass += b[j][u][0] * direction  # mass
                li += b[j][u][1] * direction  # li

            # update masses
            a[j][0][i] = a[j][0][i - 1] + mass * dt  # mass
            a[j][1][i] = a[j][1][i - 1] + li * dt  # li
            a[j][2][i] = a[j][0][i] / a[j][3][i]  # c

        i = i + 1


def build_vr_list(lvr: list) -> tuple:
    """Build lists which contain all function references for
    virtual reservoirs as well aas their input values

    """

    fn = List()  # List() # list of functions
    input_data = List()  # reservoir data
    vr_data = List()  # flux data  flux.m flux.l, flux.h, flux.d
    vr_params = List()  # list of constants
    fn = numba.typed.List.empty_list(
        types.UniTuple(types.float64, 4)(
            types.int64,  # i
            types.ListType(types.float64[::1]),
            types.ListType(types.float64[::1]),
            types.ListType(types.float64),  # a3
        ).as_type()
    )

    count = 0
    for r in lvr:  # loop over reservoir processes

        func_name, in_d, vr_d, params = r.get_process_args()
        fn.append(func_name)
        input_data.append(in_d)
        vr_data.append(vr_d)
        vr_params.append(params)
        count = count + 1

    return fn, input_data, vr_data, vr_params, count


def build_flux_lists_all(lor, iso: bool = False) -> tuple:
    """flux_list :list [] contains all fluxes as
    [f.m, f.l, f.h, f.d], where each sublist relates to one reservoir

    i.e. per reservoir we have list [f1, f2, f3], where fi = [m, l]
    and m & l  = np.array()

    iso = False/True

    """

    r_list: list = List()
    v_list: list = List()
    r0_list: list = List()
    dir_list: list = List()
    rd_list: list = List()
    f_list: List = List()

    for r in lor:  # loop over all reservoirs
        if len(r.lof) > 0:
            rd_list = List([r.m, r.l, r.c, r.v])

            r_list.append(rd_list)
            v_list.append(float(r.volume))
            r0_list.append(float(r.sp.r))

            i = 0
            # add fluxes for each reservoir entry
            tf: list = List()  # temp list for flux data
            td: list = List()  # temp list for direction data

            # loop over all fluxes
            for f in r.lof:
                tf.append(f.fa)
                td.append(float(r.lodir[i]))
                i = i + 1

            f_list.append(tf)
            dir_list.append(td)

    return r_list, f_list, dir_list, v_list, r0_list


def build_process_list(lor: list, lop: list) -> tuple:
    from numba.typed import List
    import numba
    from numba.core import types

    fn = List()  # List() # list of functions
    da = List()  # data
    pc = List()  # list of constants

    print(f"Building Process List")

    tfn = numba.typed.List.empty_list(
        types.ListType(types.void)(  # return value
            types.ListType(types.float64[::1]),  # data array
            types.ListType(types.float64),  # parameter list
            types.int64,  # parameter 4
        ).as_type()
    )

    tda = List()  # temp list for data
    tpc = List()  # temp list for constants

    # note that types.List is differenfr from Types.ListType. Also
    # note that [::1]  declares C-style arrays see
    # https://numba.discourse.group/t/list-mistaken-as-list-when-creating-list-of-function-references/677/3

    for p in lop:  # loop over reservoir processes
        # print(f"working on {p.name}")
        func_name, data, proc_const = p.get_process_args()
        tfn.append(func_name)
        tda.append(data)
        tpc.append(proc_const)

    return tfn, tda, tpc
